﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include <math.h>
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Conf.h"
#include "Application.h"
#include "Board.h"
#include "Storing.h"
#include "Cy4Link.h"
#include "Timer.h"
#include "DateTime.h"
#include "RingBuffer.h"
#include "Logger.h"
#include "BMP085.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum
{   TREND_FALLING  =   -1,
    TREND_CONSTANT =    0,
    TREND_RISING   =    1,
    TREND_UNKNOWN  = -128
} Trend;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static  Time   lastLogTime       = TIME_INVALID;
static  Timer  timeQueryInterval = TIMER_SECONDS(0);
static  Trend  airPressureTrend  = TREND_UNKNOWN;
static  Trend  temperatureTrend  = TREND_UNKNOWN;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
static SInt16
Fixp1ToInt(SInt16 value)
{
    SInt16 quotient  = value/10;
    SInt08 remainder = value%10;
    if      (remainder >  4) quotient++;
    else if (remainder < -4) quotient--;
    return quotient;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  PressureAtNN(): Umrechnung absoluter Druck auf relativen Druck (Druck bezogen auf   //
//  --------------  Meereshöhe). Grundlage ist eine Variante der internationalen Höhen- //
//  formel gemäß Datenblatt zum Barometersensor BMP085 von Bosch, Seite 14.             //
//                                                                                      //
//                  Beispiel: PressureAtNN(110, 990) -> 1003 // hPa                     //
//                                                                                      //
//  Eingabe: · altitude: Höhe über NN, -300 .. 7000 [m].                                //
//           · pressure: Absoluter Druck in 'altitude', 300 .. 1100 [hPa].              //
//  Ausgabe: Der relative Druck bezogen auf Meereshöhe, 300 .. 3000 [hPa] oder 0x8000.  //
//////////////////////////////////////////////////////////////////////////////////////////
static SInt16
PressureAtNN(SInt16 altitude, SInt16 pressure)
{
    if (altitude < -300 || altitude > 7000) return INVALID;
    if (pressure <  300 || pressure > 1100) return INVALID;
    return (SInt16)( (double)pressure / pow(1-(double)altitude/44330.0, 5.255) + 0.5 );
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt16
AbsAirPressure(void)
{
    return BMP085_AirAttributes()->pressure;
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt16
RelAirPressure(void)
{
    SInt16 pressure = AbsAirPressure();
    SInt16  altitude;
    STORING_Load(STO_BarometerAltitude, &altitude);
    return PressureAtNN(altitude, pressure);
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt08
MinimumTemperature(Time* time)
{
    SInt08 minValue = INT8_MAX;
    SInt16 minIndex = INVALID;

    // Minimum bestimmen, Wert und wo
    SInt16 index = 0;
    while (index < (SInt16)LOGGER_Capacity())
    {   SInt08 value = LOGGER_At(index);
        index++;
        if (value == CFG_TEMPERATURE_INVALID) continue;
        if (minValue > value || minIndex == INVALID)
        {   minValue = value;
            UInt16 plateau = 0;
            while (index+plateau < LOGGER_Capacity())
            {   if (LOGGER_At(index+plateau) != minValue) break;
                plateau++;
            }
            minIndex = index+plateau/2-1;
            index += plateau;
        }
    }
    // Zeitpunkt bestimmen
    time->hour   = 255;
    time->minute = 255;
    time->second = 255;
    if (minIndex != INVALID)
    {   *time = lastLogTime;
        UInt16 minutes = minIndex*CFG_TEMPERATURE_LOG_INTERVAL;
        TIME_ShiftHour(time,   -(SInt08)(minutes/60));
        TIME_ShiftMinute(time, -(SInt08)(minutes%60));
    }
    return minValue;
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt08
MaximumTemperature(Time* time)
{
    SInt08 maxValue = INT8_MIN;
    SInt16 maxIndex = INVALID;

    // Maximum bestimmen, Wert und wo
    SInt16 index = 0;
    while (index < (SInt16)LOGGER_Capacity())
    {   SInt08 value = LOGGER_At(index);
        index++;
        if (value == CFG_TEMPERATURE_INVALID) continue;
        if (maxValue < value || maxIndex == INVALID)
        {   maxValue = value;
            UInt16 plateau = 0;
            while (index+plateau < LOGGER_Capacity())
            {   if (LOGGER_At(index+plateau) != maxValue) break;
                plateau++;
            }
            maxIndex = index+plateau/2-1;
            index += plateau;
        }
    }
    // Zeitpunkt bestimmen
    time->hour   = 255;
    time->minute = 255;
    time->second = 255;
    if (maxIndex != INVALID)
    {   *time = lastLogTime;
        UInt16 minutes = maxIndex*CFG_TEMPERATURE_LOG_INTERVAL;
        TIME_ShiftHour(time,   -(SInt08)(minutes/60));
        TIME_ShiftMinute(time, -(SInt08)(minutes%60));
    }
    return maxValue;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichtenausstatter  ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupTemperatureBroadcast(Cy4Message* message)
{
    UInt08 data[1] = { CFG_TEMPERATURE_INVALID };
    SInt16 temperature = BOARD_Temperature();
    if (temperature != INVALID) data[0] = Fixp1ToInt(temperature);
    Cy4Address address = CY4ADDRESS_Create(0, 0);
    CY4MESSAGE_SetupWrite(message, address, 0xE00030, data, CARDINALITY(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupAbsAirPressureBroadcast(Cy4Message* message)
{
    UInt08 data[2];
    SInt16 pressure = AbsAirPressure();
    data[0] = HiByte(pressure);
    data[1] = LoByte(pressure);
    Cy4Address address = CY4ADDRESS_Create(0, 0);
    CY4MESSAGE_SetupWrite(message, address, 0xE00040, data, CARDINALITY(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupRelAirPressureBroadcast(Cy4Message* message)
{
    UInt08 data[2];
    SInt16 pressure = RelAirPressure();
    data[0] = HiByte(pressure);
    data[1] = LoByte(pressure);
    Cy4Address address = CY4ADDRESS_Create(0, 0);
    CY4MESSAGE_SetupWrite(message, address, 0xE00050, data, CARDINALITY(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupQueryTimeMessage(Cy4Message* message)
{
    Cy4Link timeServer;
    STORING_Load(STO_TimeServer, &timeServer);
    if (CY4LINK_IsValid(&timeServer))
    {   CY4MESSAGE_SetupRead(message, timeServer.address, timeServer.location, 7);
    }
    else
    {   CY4MESSAGE_Invalidate(message);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Antwort-Behandler  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ResponseOfQueryTime(const Cy4Message* message)
{
    if (CY4MESSAGE_IsValid(message))
    {   DateTime dateTime;
        DATETIME_FromData(CY4MESSAGE_ConstData(message), &dateTime);
        if (DATETIME_IsValid(&dateTime)) BOARD_SetClock(&dateTime);
    }    
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Anfrage-Bearbeiter  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: Temperature (read only)                                                     //
//                                                                                      //
//  Beschreibung: Aktuelle Temperatur und Extremwerte der letzten 24h mit Zeitpunkten.  //
//                                                                                      //
//              0   1   2   3   4   5   6   7                                           //
//            +---+---+---+---+---+---+---+---+                                         //
//  Belegung: |AT0|AT1|MIN|THH|TMM|MAX|HHH|HMM|                                         //
//            +---+---+---+---+---+---+---+---+                                         //
//                                                                                      //
//  [ATx]  Aktuelle Temperatur:              -497 .. +600 [°C÷10] (-32768)              //
//  [MIN]  Tiefsttemperatur der letzten 24h: -49   .. +60 [°C]    (-128)                //
//  [THH]  Stunde der Tiefsttemperatur:        0   ..  23 [std]   ( 255)                //
//  [TMM]  Minute der Tiefsttemperatur:        0   ..  59 [min]   ( 255)                //
//  [MAX]  Höchsttemperatur der letzten 24h: -49   .. +60 [°C]    (-128)                //
//  [HHH]  Stunde der Höchsttemperatur:        0   ..  23 [std]   ( 255)                //
//  [HMM]  Minute der Höchsttemperatur:        0   ..  59 [min]   ( 255)                //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnTemperature(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[8] = { HiByte(INVALID), LoByte(INVALID),
                        CFG_TEMPERATURE_INVALID, 255, 255,
                        CFG_TEMPERATURE_INVALID, 255, 255
                      };

    SInt16 temperature = BOARD_Temperature();
    data[0] = HiByte(temperature);
    data[1] = LoByte(temperature);
    if (TIME_IsValid(&lastLogTime))
    {   Time time;
        SInt08 temperatureMin = MinimumTemperature(&time);
        if (temperatureMin != CFG_TEMPERATURE_INVALID)
        {   data[2] = temperatureMin;
            data[3] = time.hour;
            data[4] = time.minute;
        }
        SInt08 temperatureMax = MaximumTemperature(&time);
        if (temperatureMax != CFG_TEMPERATURE_INVALID)
        {   data[5] = temperatureMax;
            data[6] = time.hour;
            data[7] = time.minute;
        }
    }
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ThermistorAdcRawValue (read only)                                           //
//                                                                                      //
//  Beschreibung: Direkter ADC-Abtastwert des Thermistors.                              //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |SPH|SPL|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [SPx]  Der ADC-Abtastwert des Thermistors: 0 .. 1023. [#]                           //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnThermistorAdcRawValue(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[2];

    UInt16 sample = BOARD_ReadThermistor();
    data[0] = HiByte(sample);
    data[1] = LoByte(sample);
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: TemperatureTrend (read only)                                                //
//                                                                                      //
//  Beschreibung: Trend der Temperaturentwicklung über die letzten 2 Stunden.           //
//                Mögliche Werte sind:                                                  //
//                                                                                      //
//                             -1 = Fallend,                                            //
//                              0 = Gleichbleibend                                      //
//                              1 = Steigend                                            //
//                           -128 = Unbekannt                                           //
//                                                                                      //
//            +---+                                                                     //
//  Belegung: |TND|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [TND]  Temperatur-Trend: -1, 0, 1, -128.                                            //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnTemperatureTrend(const Cy4Service* service, Cy4Message* message)
{
	(void)service;
    CY4MESSAGE_Data(message)[0] = (UInt08)temperatureTrend;
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: TemperatureLog (read only)                                                  //
//                                                                                      //
//  Beschreibung: Temperaturaufzeichnung der letzten 24h als Mittelwerte im 15-Minuten  //
//                Raster.                                                               //
//                                                                                      //
//              0   1   2   ...  93  94  95                                             //
//            +---+---+---+-   -+---+---+---+                                           //
//  Belegung: |T00|T01|T02| ... |T93|T94|T95|                                           //
//            +---+---+---+-   -+---+---+---+                                           //
//                                                                                      //
//  [T00]  Mittelwert der Temperatur der letzten    15 Minuten: -49 .. +60 [°C] (-128)  //
//  [T01]  Mittelwert der Temperatur von vor 15 bis 30 Minuten: -49 .. +60 [°C] (-128)  //
//  [T02]  Mittelwert der Temperatur von vor 30 bis 45 Minuten: -49 .. +60 [°C] (-128)  //
//    :        :               :                                 :      :    :     :    //
//  [T93]  Mittelwert der Temp. von vor 23h 15m bis 23h 30m:    -49 .. +60 [°C] (-128)  //
//  [T94]  Mittelwert der Temp. von vor 22h 30m bis 23h 45m:    -49 .. +60 [°C] (-128)  //
//  [T95]  Mittelwert der Temp. von vor 23h 45m bis 24h 00m:    -49 .. +60 [°C] (-128)  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnTemperatureLog(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[96];

    UInt16 index = 0;
    for (UInt08 i=0; i<CARDINALITY(data); i++)
    {   SInt16 sum   = 0;                               // Mittelwertsumme
        UInt08 size  = 15/CFG_TEMPERATURE_LOG_INTERVAL; // Mittelwertbreite
        UInt08 count = 0;                               // Anzahl aufsummierter Werte
        for (UInt08 j=0; j < size; j++)
        {   if (index == LOGGER_Capacity()) break;
            SInt08 value = LOGGER_At(index);
            if (value != CFG_TEMPERATURE_INVALID)
            {   sum += value;
                count++;
            }
            index++;
        }
        if (count > size/2)
        {   // Mittelwert nur ermitteln, wenn mehr als die hälfte der Werte i.O. sind
            SInt08 average = sum/count;
            SInt08 rest    = sum%count;
            if (rest<0) rest = -rest;
            if ((count&1) == 0) rest++;
            if (rest > count/2) (sum<0) ? average-- : average++;
            data[i] = average;
        }
        else
        {   data[i] = CFG_TEMPERATURE_INVALID;
        }
    }
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AirPressureAbs (read only)                                                  //
//                                                                                      //
//  Beschreibung: Absoluter Umgebungsluftdruck.                                         //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |AP1|AP0|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [APx]  Absoluter Luftdruck: 300 .. 1100 [hPa] (0x8000)                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAirPressureAbs(const Cy4Service* service, Cy4Message* message)
{
    SInt16 pressure = AbsAirPressure();
    UInt08 data[2];
    data[0] = (UInt08)((UInt16)(pressure)>>8);
    data[1] = (UInt08)((UInt16)(pressure)>>0);
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AirPressureRel (read only)                                                  //
//                                                                                      //
//  Beschreibung: Umgebungsluftdruck bezogen auf Meeresspiegel. Vorraussetzung für die  //
//                Ermittlung des sogenannten relativen Drucks, ist die Angabe der Höhe  //
//                über NN (siehe Dienst BarometerAltitude)                              //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |RP1|RP0|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [RPx]  Relativer Luftdruck: 300 .. 1100 [hPa] (-32768)                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAirPressureRel(const Cy4Service* service, Cy4Message* message)
{
    SInt16 pressure = RelAirPressure();
    UInt08 data[2];
    data[0] = (UInt08)((UInt16)(pressure)>>8);
    data[1] = (UInt08)((UInt16)(pressure)>>0);
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: BarometerAltitude (read only)                                               //
//                                                                                      //
//  Beschreibung: Höhe des Barometers über Meeresspiegel. Die Angabe der Höhe über NN   //
//                ist die Vorraussetzung für die korrekte Ermittlung des relativen      //
//                Umgebungsluftdrucks (siehe AirPressureRel).                           //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NN1|NN0|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NNx]  Höhe über Meereshöhe: -300 .. 7000 [m] (0x8000)                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBarometerAltitude(const Cy4Service* service, Cy4Message* message)
{
    SInt16 altitude;

    STORING_Load(STO_BarometerAltitude, &altitude);
    if( altitude < -300 || altitude > 7000 ) altitude = INVALID;
    UInt08 data[2];
    data[0] = (UInt08)((UInt16)(altitude)>>8);
    data[1] = (UInt08)((UInt16)(altitude)>>0);
    CY4SERVICE_TransferPayload(service, message, data);
    if( CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE )
    {   altitude = (SInt16)((((UInt16)data[0])<<8) | data[1]);
        STORING_Save(&altitude, STO_BarometerAltitude);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: BaroTemperature (read only)                                                 //
//                                                                                      //
//  Beschreibung: Aktuelle Temperatur im Barometersensor.                               //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |TM0|TM1|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [TPx]  Aktuelle Temperatur im Barometer-Chip:  -994 .. +994 [°C÷10] oder (-32768)   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBaroTemperature(const Cy4Service* service, Cy4Message* message)
{
    SInt16 temperature = BMP085_AirAttributes()->temperature;
    UInt08 data[2] = { HiByte(temperature), LoByte(temperature) };
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AirPressureTrend (read only)                                                //
//                                                                                      //
//  Beschreibung: Tendenz der Luftdruckentwicklung über die letzten 2 Stunden. Mögliche //
//                Werte sind:                                                           //
//                             -1 = Fallend,                                            //
//                              0 = Gleichbleibend                                      //
//                              1 = Steigend                                            //
//                           -128 = Unbekannt                                           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |TND|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [TND]  Luftdruck-Trend:  -1, 0, 1, -128.                                            //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAirPressureTrend(const Cy4Service* service, Cy4Message* message)
{
	(void)service;
    CY4MESSAGE_Data(message)[0] = (UInt08)airPressureTrend;
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: BroadcastsEnable (read write)                                               //
//                                                                                      //
//  Beschreibung: Aktivierung der Rundrufe. Die Wert werden jeweils an 000.000 über-    //
//                tragen, sobald sie sich ändern, jedoch nicht häufiger als im 5-       //
//                Minuten-Intervall. Der Temperatur-Rundruf sollte nur aktiviert        //
//                werden, wenn es sich bei der erfassten Temperatur um die aktuelle     //
//                Außentemperatur handelt. Die Rundrufe werden an folgende Dienstort-   //
//                bereich gesendet:                                                     //
//                                                                                      //
//                    Außentemperatur:            E00030h:1:WO, Außentemperatur: -49 .. //
//                                                +60 [°C] (Fehler: -128)               //
//                    Umgebungsluftdruck absolut: E00040h:2:WO, Luftdruck: 300 .. 1100  //
//                                                [hPA] (Fehler: -32768)                //
//                    Umgebungsluftdruck relativ: E00050h:2:WO, Luftdruck: 300 .. 1100  //
//                                                [hPA] (Fehler: -32768)                //
//                                                                                      //
//               0   1   2                                                              //
//             +---+---+---+                                                            //
//  Belegung:  |TMP|AAP|RAP|                                                            //
//             +---+---+---+                                                            //
//                                                                                      //
//  [TMP]: Rundruf Außentemperatur: 1 = Ausführen, 0 = Nicht ausführen (Voreinstellung) //
//  [AAP]: Rundruf absoluter Umgebungsluftdruck: 1 = Ausführen, 0 = Nicht ausführen     //
//         (Voreinstellung).                                                            //
//  [RAP]: Rundruf relativer Umgebungsluftdruck: 1 = Ausführen, 0 = Nicht ausführen     //
//         (Voreinstellung).                                                            //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBroadcastsEnable(const Cy4Service* service, Cy4Message* message)
{
    bool enabled[3];
    STORING_Load(STO_BroadcastEnabled, &enabled);
    UInt08 data[3];
    data[0] = enabled[0];
    data[1] = enabled[1];
    data[2] = enabled[2];
    CY4SERVICE_TransferPayload(service, message, data);
    enabled[0] = (bool)data[0];
    enabled[1] = (bool)data[1];
    enabled[2] = (bool)data[2];
    STORING_Save(&enabled, STO_BroadcastEnabled);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: TimeServerLink (read write)                                                 //
//                                                                                      //
//  Beschreibung: Adresse und Dienstort des Zeitservers. Es werden 7 Bytes angefragt,   //
//                dazu s.a. DateTime.                                                   //
//                                                                                      //
//              0   1   2   3   4                                                       //
//            +---+---+---+---+---+                                                     //
//  Belegung: |NET|DEV|LC1|LC2|LC3|                                                     //
//            +---+---+---+---+---+                                                     //
//                                                                                      //
//  [NET] Cy4-Netzadresse des Zeitservers: 1 .. 255.                                    //
//  [DEV] Cy4-Geräteadresse des Zeitservers: 1 .. 255.                                  //
//  [LCx] Dienstort der abzufragenden Zeitinformation: 100000h .. FFFF00h, FFFFFFh.     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnTimeServerLink(const Cy4Service* service, Cy4Message* message)
{
    Cy4Link  timeServer;
    STORING_Load(STO_TimeServer, &timeServer);
    UInt08  data[5];
    CY4LINK_ToData(&timeServer, data);
    CY4SERVICE_TransferPayload(service, message, data);
    CY4LINK_FromData(data, &timeServer);
    STORING_Save(&timeServer, STO_TimeServer);
    if (CY4LINK_IsValid(&timeServer)) TIMER_Start(&timeQueryInterval, 3);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AcceptTimeBroardcast (read write)                                           //
//                                                                                      //
//  Beschreibung: Mit diesem Dienst kann festgelegt werden, ob das Gerät Zeit-Rund-     //
//                mitteilungen an 0xE00020 entgegennimmt oder nicht.                    //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |ATB|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [ATB] Zeit-Rundmitteilung: 1 = akzeptieren, 0 = ignorieren.                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAcceptTimeBroardcast(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    bool  enable;
    if ( CY4MESSAGE_Access(message) == CY4MESSAGE_READ)
    {   STORING_Load(STO_AcceptTimeBroadcast, &enable);
        CY4MESSAGE_Data(message)[0] = enable;
    }
    else
    {   enable = !!CY4MESSAGE_ConstData(message)[0];
        STORING_Save(&enable, STO_AcceptTimeBroadcast);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DateTime (read write)                                                       //
//                                                                                      //
//  Beschreibung: Ermöglicht den Zugriff auf die Systemzeit des Gerätes. Der Behandler  //
//                wird für den lokalen Dienst wie auch für den Rundruf eingesetzt.      //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |YR1|YR2|MON|DAY|HUR|MIN|SEC|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//    [YRx] Jahr:    1700 .. 9000 , (0xFFFF)                                            //
//    [MON] Monat:      1 ..   12 , (0xFF)                                              //
//    [DAY] Tag:        1 ..   31 , (0xFF)                                              //
//    [HUR] Stunde:     0 ..   23 , (0xFF)                                              //
//    [MIN] Minute:     0 ..   59 , (0xFF)                                              //
//    [SEC] Sekunde:    0 ..   59 , (0xFF)                                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDateTime(const Cy4Service* service, Cy4Message* message)
{
    if (CY4MESSAGE_Location(message) == 0xE00020)
    {   bool broadcastsAccepted;
        STORING_Load(STO_AcceptTimeBroadcast, &broadcastsAccepted);
        if (!broadcastsAccepted) return;
    }
    UInt08 data[7];
    {   DateTime  dateTime;
        BOARD_Clock(&dateTime);
        DATETIME_ToData(&dateTime, data);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_WRITE)
    {   DateTime dateTime;
        DATETIME_FromData(data, &dateTime);
        BOARD_SetClock(&dateTime);
        TIMER_Start(&timeQueryInterval, CFG_TIME_QUERY_INTERVAL*60);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: InfoLedSignal (read write)                                                  //
//                                                                                      //
//  Beschreibung: Legt die Aufgabe der grünen Info-LEDs fest. Folgende Lichtsignale     //
//                können eingestellt werden:                                            //
//                                                                                      //
//                       0 = Kein Signal, Ausgeschaltet.                                //
//                       1 = Leuchtet durchgehend.                                      //
//                       2 = Herzschlag-Signal.                                         //
//                       3 = Zeigt Aktivität auf dem Bus.                               //
//                       x = Wirkungslos.                                               //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |SIG|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [SIG] Lichtsignale der grünen Info-LED, Voreingestellt ist 3.                       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnInfoLedSignal(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[1];
    {   InfoLedSignal signal;
        STORING_Load(STO_InfoLedSignal, &signal);
        data[0] = (UInt08)signal;
    }
    CY4SERVICE_TransferPayload(service, message, data);
    if (data[0] < 4)
    {   InfoLedSignal signal = (InfoLedSignal)data[0];
        STORING_Save(&signal, STO_InfoLedSignal);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N01|N02|N03| ... |N60| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[sizeof(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand.           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 5Eh ('^').          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) STORING_RestoreDefaults();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//                                                                                      //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
    CY4TYPEPLATE(DEV_APPL_TYPE,
    DEV_APPL_VERS,
    DEV_BOARD_TYPE)
};

UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
{   CY4MESSAGE_Data(message)[i] = pgm_read_byte(&typePlate[1][j]);
    j++;
}
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 5Eh ('^').                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 5Eh ('^').                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeCy4Address (write only)                                        //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeCy4Address(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAdr(&address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedTemperatureLogging(void)
{
    static Timer update = TIMER_SECONDS(CFG_TEMPERATURE_LOG_INTERVAL*60);

    if( TIMER_RestartIfExpired(&update) )
    {   DateTime dateTime;
        BOARD_Clock(&dateTime);
        if( DATETIME_IsValid(&dateTime) )
        {   lastLogTime = dateTime.time;
            SInt16 temperature = BOARD_Temperature();
            if (temperature == INVALID) LOGGER_Log(CFG_TEMPERATURE_INVALID);
            else                        LOGGER_Log(Fixp1ToInt(temperature));
        }
        else
        {   lastLogTime.hour   = 255;
            lastLogTime.minute = 255;
            lastLogTime.second = 255;
            LOGGER_Clear();
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ProceedTemperatureTrend(void)
{
    static SInt16      buffer[25] = { INVALID };
    static RingBuffer  ringBuffer = RINGBUFFER_INIT(buffer);
    static Timer       update     = TIMER_SECONDS(CFG_AIR_PRESSURE_TREND_PERIOD*60/
                                                  CARDINALITY(buffer));

    if (!TIMER_RestartIfExpired(&update)) return;
    SInt16 currentValue = BOARD_Temperature();  // [°C÷10]
    RINGBUFFER_Add(&ringBuffer, currentValue);
    SInt16 trend = RINGBUFFER_Trend(&ringBuffer, CFG_AIR_PRESSURE_TREND_PERIOD);
    if (trend != INVALID)
    {   if      (trend >  5) temperatureTrend = TREND_RISING;
        else if (trend < -5) temperatureTrend = TREND_FALLING;
        else                 temperatureTrend = TREND_CONSTANT;
    }
    else
    {   temperatureTrend = TREND_UNKNOWN;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ProceedAirPressureTrend(void)
{
    static SInt16      buffer[25] = { INVALID };
    static RingBuffer  ringBuffer = RINGBUFFER_INIT(buffer);
    static Timer       update     = TIMER_SECONDS(CFG_AIR_PRESSURE_TREND_PERIOD*60/
                                                  CARDINALITY(buffer)+5);

    if (!TIMER_RestartIfExpired(&update)) return;
    SInt16 currentValue = AbsAirPressure();
    RINGBUFFER_Add(&ringBuffer, currentValue);
    SInt16 trend = RINGBUFFER_Trend(&ringBuffer, CFG_AIR_PRESSURE_TREND_PERIOD);
    if (trend != INVALID)
    {   if      (trend > 0) airPressureTrend = TREND_RISING;
        else if (trend < 0) airPressureTrend = TREND_FALLING;
        else                airPressureTrend = TREND_CONSTANT;
    }
    else
    {   airPressureTrend = TREND_UNKNOWN;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ProceedBroadcasts(void)
{
    bool broadcastEnabled[3];
    STORING_Load(STO_BroadcastEnabled, &broadcastEnabled);
    if (broadcastEnabled[0])
    {   static Timer updatePause = TIMER_SECONDS(CFG_MIN_BROADCAST_PAUSE+0);
        if (TIMER_HasExpired(&updatePause))
        {   static SInt08 old = CFG_TEMPERATURE_INVALID;
            SInt08 new = CFG_TEMPERATURE_INVALID;
            SInt16 temperature = BOARD_Temperature();
            if (temperature != INVALID) new = Fixp1ToInt(temperature);
            if (old != new)
            {   old  = new;
                TIMER_RestartIfExpired(&updatePause);
                CY4NET_Post(SetupTemperatureBroadcast, NULL);
            }
        }
    }
    else if(broadcastEnabled[1])
    {   static Timer updatePause = TIMER_SECONDS(CFG_MIN_BROADCAST_PAUSE+2);
        if (TIMER_HasExpired(&updatePause))
        {   static SInt16 old = INVALID;
            SInt08 new = AbsAirPressure();
            if (old != new)
            {   old  = new;
                TIMER_RestartIfExpired(&updatePause);
                CY4NET_Post(SetupAbsAirPressureBroadcast, NULL);
            }
        }
    }
    else if(broadcastEnabled[2])
    {   static Timer updatePause = TIMER_SECONDS(CFG_MIN_BROADCAST_PAUSE+4);
        if (TIMER_HasExpired(&updatePause))
        {   static SInt16 old = INVALID;
            SInt08 new = RelAirPressure();
            if (old != new)
            {   old  = new;
                TIMER_RestartIfExpired(&updatePause);
                CY4NET_Post(SetupRelAirPressureBroadcast, NULL);
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProccedTimeRequests(void)
{
    if (TIMER_HasExpired(&timeQueryInterval))
    {   TIMER_Start(&timeQueryInterval, CFG_TIME_QUERY_INTERVAL*60);
        CY4NET_Post(SetupQueryTimeMessage, ResponseOfQueryTime);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedLedSignaling(void)
{
    InfoLedSignal signal;
    STORING_Load(STO_InfoLedSignal, &signal);
    UInt16 ticks = BOARD_MSecTicks();
    bool failure = false;
    // Fehlerprüfung (erst 30 s nach Systemstart beginnen)
    {   static Timer failureCheckDelay = TIMER_SECONDS(0);
        if (TIMER_Duration(&failureCheckDelay) == 0)
        {   TIMER_Start(&failureCheckDelay, 30);
        }
        else if (TIMER_HasExpired(&failureCheckDelay))
        {   failure = BOARD_Temperature() == INVALID ||
            BMP085_AirAttributes()->temperature == INVALID ||
            AbsAirPressure() == INVALID;
        }
    }
    if (!failure)
    {   bool ledOn;
        switch (signal)
        {   case INFOLEDSIGNAL_OFF:       ledOn = false;                  break;
            case INFOLEDSIGNAL_ON:        ledOn = true;                   break;
            case INFOLEDSIGNAL_HEARTBEAT: ledOn = (bool)((ticks & 0x380)==0x80 || (ticks & 0x380)==0x180); break;
            case INFOLEDSIGNAL_BUS_ACT:   ledOn = BOARD_BusActivity();    break;
            default:                      ledOn = false;
        }
        BOARD_SetLed(LED_GREEN, ledOn);
        BOARD_SetLed(LED_RED  , false);
    }
    else
    {   // Fehler: Rote LED blinken lassen
        bool ledOn;
        ledOn = (bool)(ticks & 0x0100);
        BOARD_SetLed(LED_RED  , ledOn);
        ledOn = signal == INFOLEDSIGNAL_BUS_ACT && BOARD_BusActivity();
        BOARD_SetLed(LED_GREEN, ledOn);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                 service                  |         handler         |
        //+----------+------+------------------------+-------------------------+
        { { 0x100100 ,   8  , CY4SERVICE_READ_ONLY  }, OnTemperature           },
        { { 0x10010C ,   2  , CY4SERVICE_READ_ONLY  }, OnThermistorAdcRawValue },
        { { 0x100110 ,   1  , CY4SERVICE_READ_ONLY  }, OnTemperatureTrend      },
        { { 0x100120 ,  96  , CY4SERVICE_READ_ONLY  }, OnTemperatureLog        },
        { { 0x100200 ,   2  , CY4SERVICE_READ_ONLY  }, OnAirPressureAbs        },
        { { 0x100210 ,   2  , CY4SERVICE_READ_ONLY  }, OnAirPressureRel        },
        { { 0x100220 ,   2  , CY4SERVICE_READ_WRITE }, OnBarometerAltitude     },
        { { 0x100230 ,   2  , CY4SERVICE_READ_ONLY  }, OnBaroTemperature       },
        { { 0x100240 ,   1  , CY4SERVICE_READ_ONLY  }, OnAirPressureTrend      },
        { { 0x100300 ,   3  , CY4SERVICE_READ_WRITE }, OnBroadcastsEnable      },
        { { 0x100410 ,   7  , CY4SERVICE_READ_WRITE }, OnDateTime              },
        { { 0x100430 ,   5  , CY4SERVICE_READ_WRITE }, OnTimeServerLink        },
        { { 0x100440 ,   1  , CY4SERVICE_READ_WRITE }, OnAcceptTimeBroardcast  },
        { { 0xE00020 ,   7  , CY4SERVICE_WRITE_ONLY }, OnDateTime              },
        { { 0xFFF020 ,   1  , CY4SERVICE_READ_WRITE }, OnInfoLedSignal         },
        { { 0xFFFF00 ,  61  , CY4SERVICE_READ_WRITE }, OnDeviceLabel           },
        { { 0xFFFF40 ,   1  , CY4SERVICE_WRITE_ONLY }, OnFactorySettings       },
        { { 0xFFFFF0 ,   7  , CY4SERVICE_READ_ONLY  }, OnCallTypePlate         },
        { { 0xFFFFF8 ,   1  , CY4SERVICE_WRITE_ONLY }, OnEnterApplication      },
        { { 0xFFFFFA ,   1  , CY4SERVICE_WRITE_ONLY }, OnEnterProgramLoader    },
        { { 0xFFFFFC ,   2  , CY4SERVICE_WRITE_ONLY }, OnChangeCy4Address}
    };    
    CY4NET_AssignServices(servicesTable);
    CY4NET_SetAddress(STORING_LoadCy4SelfAdr());
    LOGGER_Clear();
    TIMER_Start(&timeQueryInterval, CFG_INITIAL_TIME_QUERY_DELAY);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
    ProceedTemperatureLogging();
    ProceedTemperatureTrend();
    ProceedAirPressureTrend();
    ProceedBroadcasts();
    ProccedTimeRequests();
    ProceedLedSignaling();
    BMP085_Proceed();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
