﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Umgebungsbedingungs-Sensor BMP085 Bosch                                             //
//  ________________________________________________________________________            //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon												    //
//                                                                                      //
//  Grundlage der Software sind die Kompensationsmethoden aus dem BMP085-Datenblatt und //
//  des BMP085-API der Bosch Sensortec GmbH, erhältlich unter www.bosch-sensortec.com.  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "Conf.h"
#include "Board.h"
#include "Timer.h"
#include "BMP085.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define I2C_ADR 0x77

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static  AirAttributes  airAttributes = { INVALID, INVALID };

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
const AirAttributes*
BMP085_AirAttributes(void)
{
    return &airAttributes;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BMP085_Proceed: Liest den statischen Absolut-Luftdruck und die Temperatur des Um-	//
//  --------------  gebunsdrucksensors vom vom Typ BMP085 aus (siehe dazu Bosch Daten-  //
//  blatt Rev. 1.2 bzw. 2.5 S. 15 ff.). Die Werte werden in AirAttributes angelegt.		//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BMP085_Proceed(void)
{
	enum    		// Register:    Typ:
	{   AC1 =  0,   // 0xAA,0xAB    SInt16
		AC2 =  1,   // 0xAC,0xAD    SInt16
		AC3 =  2,   // 0xAE,0xAF    SInt16
		AC4 =  3,   // 0xB0,0xB1    UInt16
		AC5 =  4,   // 0xB2,0xB3    UInt16
		AC6 =  5,   // 0xB4,0xB5    UInt16
		B1  =  6,   // 0xB6,0xB7    SInt16
		B2  =  7,   // 0xB8,0xB9    SInt16
		MB  =  8,   // 0xBA,0xBB    SInt16
		MC  =  9,   // 0xBC,0xBD    SInt16
		MD  = 10    // 0xBE,0xBF    SInt16
	};

	typedef enum
	{   READ_CAL,
		SEL_TEMP,
		READ_TEMP,
		SEL_PRESS,
		READ_PRESS
	} Action;
	
	const   UInt08  OSS     = 3;      // Oversampling: 0..3
	static  Timer   timer   = TIMER_MILLISECONDS(100);
	static  Action  action  = READ_CAL;
	static  SInt16  cal[]   = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	static  UInt16  uTemp   = 0xFFFF;
	UInt08          i;
	UInt08          hi;
	UInt08          lo;
	UInt08          data[3];
	SInt32          up;
	SInt32          x1;
	SInt32          x2;
	SInt32          x3;
	SInt32          b3;
	SInt32          b5;
	SInt32          b6;
	SInt32          p;
	UInt32          b4;
	UInt32          b7;

	if (!TIMER_HasExpired(&timer)) return;
	TIMER_Start(&timer, 100);
	switch (action)
	{   case READ_CAL:
		data[0] = 0xAA;
		if (BOARD_I2cTransmit(I2C_ADR, data, 1))
		{   if (BOARD_I2cReceive(I2C_ADR, (UInt08*)cal, CARDINALITY(cal)*2))
			{   for (i=0; i < CARDINALITY(cal); i++)
				{   hi     = ((const UInt08*)cal)[2*i  ];
					lo     = ((const UInt08*)cal)[2*i+1];
					cal[i] = ((((UInt16)hi)<<8) | (UInt16)lo);
					if (cal[i] == 0 || cal[i] == (SInt16)0xFFFF) break;
				}
				if (i == CARDINALITY(cal)) action = SEL_TEMP;
			}
		}
		break;
		case SEL_TEMP:
		uTemp = 0xFFFF;
		data[0] = 0xF4;
		data[1] = 0x2E;
		if (BOARD_I2cTransmit(I2C_ADR, data, 2)) action = READ_TEMP;
		else                                  action = SEL_PRESS;
		break;
		case READ_TEMP:
		uTemp = 0xFFFF;
		data[0] = 0xF6;
		if (BOARD_I2cTransmit(I2C_ADR, data, 1) && BOARD_I2cReceive(I2C_ADR, data, 2))
		{   uTemp = (SInt16)(((UInt16)data[0]<<8) | (UInt16)data[1]);
		}
		action = SEL_PRESS;
		break;
		case SEL_PRESS:
		if (uTemp == 0xFFFF)
		{   airAttributes.pressure = INVALID;
			action = SEL_TEMP;
		}
		else
		{   data[0] = 0xF4;
			data[1] = 0x34 + (OSS << 6);
			if (BOARD_I2cTransmit(I2C_ADR, data, 2)) action = READ_PRESS;
			else                                  action = SEL_TEMP;
		}
		break;
		case READ_PRESS:
		airAttributes.pressure = INVALID;
		data[0] = 0xF6;
		if (BOARD_I2cTransmit(I2C_ADR, data, 1) && BOARD_I2cReceive(I2C_ADR, data, 3))
		{   up = (SInt32)( ((UInt32)data[0]<<16) |
			((UInt32)data[1]<< 8) |
			((UInt32)data[2]<< 0) );
			up >>= 8 - OSS;
			// Berechnung: Wirklicher Temperaturwert
			x1 = (((SInt32)uTemp - (UInt16)cal[AC6]) * (UInt16)cal[AC5]) >> 15;
			x2 = (((SInt32)cal[MC])<<11)/(x1 + cal[MD]);
			b5 = x1 + x2;
			{   // Erfassung der Sensor-Chiptemperatur: -994 .. +994 [°C÷10]
				SInt32 t = (b5 + 8) >> 4;
				if (t > -995 && t < 995) airAttributes.temperature = (SInt16)t;
				else                     airAttributes.temperature = INVALID;
			}
			// Berechnung: Wirklicher Druckwert
			b6 = b5 - 4000;
			x1 = (cal[B2] * (b6*b6 >> 12)) >> 11;
			x2 = cal[AC2] * b6 >> 11;
			x3 = x1 + x2;
			b3 = ((((SInt32)cal[AC1]*4 + x3)<<OSS)+2) >> 2;
			x1 = cal[AC3]*b6 >> 13;
			x2 = (cal[B1] * (b6*b6 >> 12)) >> 16;
			x3 = ((x1+x2)+2) >> 2;
			b4 = ((UInt16)cal[AC4] * (UInt32)(x3+32768)) >> 15;
			b7 = ((UInt32)up - b3) * (50000 >> OSS);
			p  = b7 < 0x80000000 ? (b7 * 2) / b4 : (b7 / b4) * 2;
			x1 = (p >> 8) * (p >> 8);
			x1 = (x1 * 3038) >> 16;
			x2 = (-7357 * p) >> 16;
			airAttributes.pressure = (SInt16)((p + ((x1 + x2 + 3791) >> 4))/100);
			TIMER_Start(&timer, CFG_AIR_PRESSURE_UPDATE*1000);
		}
		action = SEL_TEMP;
		break;
		default:
		action = READ_CAL;
	}
}


//////////////////////////////////////////////////////////////////////////////////////////
