﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Ringpuffer (für SInt16 Werte)                                                       //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2016 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Beispiel Instanziierung:                                                            //
//                                                                                      //
//      static SInt16 buffer[RINGBUFFER_CAPACITY];  // RINGBUFFER_CAPACITY: 1 .. 255    //
//      static RingBuffer ringBuffer = RINGBUFFER_INIT(buffer);                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "RingBuffer.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen   /////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  RINGBUFFER_Size(): Liefert die Größe des Ringpuffers. Die Größe entspricht der An-  //
//  -----------------  zahl an Einträgen. Sie wird bei der Instanziierung festgelegt    //
//  und danach nicht mehr verändert.                                                    //
//                                                                                      //
//  Eingabe: · ringBuffer: Der Ringpuffer.                                              //
//  Ausgabe: Größe des Ringpuffers.                                                     //
//////////////////////////////////////////////////////////////////////////////////////////
UInt08
RINGBUFFER_Size(const RingBuffer* ringBuffer)
{
    return ringBuffer->size;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  RINGBUFFER_Fill(): Überschreibt alle Einträge des Ringpuffer mit <value>.           //
//  -----------------                                                                   //
//                                                                                      //
//  Eingabe: · ringBuffer: Der Ringpuffer.                                              //
//           · value:      Der neue Wert aller Ringpuffer-Einträge.                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
RINGBUFFER_Fill(RingBuffer* ringBuffer, SInt16 value)
{
    for (UInt08 i=0; i<ringBuffer->size; i++)
    {   ringBuffer->buffer[i] = value;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  RINGBUFFER_Add(): Fügt einen neuen Eintrag in den Ringpuffer ein und verdrängt      //
//  ----------------  damit den ältesten, d.h. den am weitesten zurückliegenden. Die    //
//  Position des neuen Eintrags ist 0.                                                  //
//                                                                                      //
//  Eingabe: · ringBuffer: Der Ringpuffer.                                              //
//           · value:      Der neu aufzunehmende Eintrag.                               //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
RINGBUFFER_Add(RingBuffer* ringBuffer, SInt16 value)
{
    UInt08 start = ringBuffer->recent;
    if (start == 0) start = ringBuffer->size-1;
    else            start--;
    ringBuffer->buffer[start] = value;
    ringBuffer->recent = start;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  RINGBUFFER_At(): Liefert den Eintrag des Ringpuffer an Position <position>. Der     //
//  ---------------  Eintrag an Position 0 ist der zuletzt eingefügte, also neueste     //
//  Eintrag. Liegt  <position> außerhalb der erlaubten Grenzen, ist das Verhalten der   //
//  Funktion undefiniert.                                                               //
//                                                                                      //
//  Eingabe: · ringBuffer: Der Ringpuffer.                                              //
//           · position:   Position des zu lesenden Eintrags: 0 .. RINGBUFFER_Size()-1. //
//  Ausgabe: Ringpuffereintrag an Position <position>.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
SInt16
RINGBUFFER_At(const RingBuffer* ringBuffer, UInt08 position)
{
    UInt08 index = ringBuffer->recent + position;
    if      (index >= ringBuffer->size) index -= ringBuffer->size;
    else if (index <  ringBuffer->recent) index += 256 - ringBuffer->size;
    return ringBuffer->buffer[index];
}

//////////////////////////////////////////////////////////////////////////////////////////
//  RINGBUFFER_Average(): Liefert den arithmetischen Mittelwert aller Ringpuffer-Ein-   //
//  --------------------  träge. Kommt als Eintrag im Ringpuffers der Wert 'INVALID'    //
//  vor, ist der Mittelwert ebenfalls 'INVALID'.                                        //
//                                                                                      //
//  Eingabe: · ringBuffer: Der Ringpuffer.                                              //
//  Ausgabe: Arithmetischer Mittelwert oder 'INVALID'.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
SInt16
RINGBUFFER_Average(const RingBuffer* ringBuffer)
{
    SInt32 sum = 0;
    UInt08 size = ringBuffer->size;
    for (UInt08 i=0; i<size; i++)
    {   SInt16 entry = ringBuffer->buffer[i];
        if (entry == INVALID) return INVALID;
        sum += entry;
    }
    SInt16 average = (SInt16)(sum/size);
    SInt16 rest    = sum%size;
    if (rest<0) rest = -rest;
    if ((size&1) == 0) rest++;
    if (rest > size/2) (sum<0) ? average-- : average++;
    return average;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  RINGBUFFER_Trend(): Liefert die Trend der Werteenwicklung im Ringpuffer. Ausge-     //
//  ------------------  geben wird die prozentuale Veränderung der Werte im Bereich     //
//  von ±1000%. Basis der Ermittlung ist die Steigung der linearen Regressionsgeraden.  //
//  Kommt als Eintrag im Ringpuffers der Wert 'INVALID' vor, ist der Trend ebenfalls    //
//  'INVALID'. Der Ringpuffer darf zwischen 5 und 100 Einträge haben.                   //
//                                                                                      //
//  Eingabe: · ringBuffer: Der Ringpuffer.                                              //
//           · period:     Zeitspanne des Ringpuffers: 1 .. 2500.                       //
//  Ausgabe: Veränderung in Prozent: -1000 .. +1000 [%] oder INVALID.                   //
//////////////////////////////////////////////////////////////////////////////////////////
SInt16
RINGBUFFER_Trend(const RingBuffer* ringBuffer, SInt16 period)
{
    UInt08 n = RINGBUFFER_Size(ringBuffer); // 3 .. 100
    // Arithmetisches Mittel der xᵢ, da xᵢ={0,..,n-1}: avX = ½(n²-n)÷n = ½(n-1)
    SInt08 avX = (n-1)/2;
    // Arithmetisches Mittel der yᵢ, avY = ∑(yᵢ)÷n
    SInt16 avY;
    {   avY = RINGBUFFER_Average(ringBuffer);
        if (avY == -32768 ) return -32768;
    }
    // w = ∑((xᵢ-avX)·(yᵢ-avY))
    SInt64 w = 0;
    {   for (UInt08 i=0; i<n; i++)
        {   UInt08 xi = i;
            SInt16 yi = RINGBUFFER_At(ringBuffer, i);
            w += (xi-avX) * (SInt64)(yi-avY);
        }
    }
    // q = ∑(xᵢ-avX)²
    SInt32 q = 0;
    {   for (UInt08 i=0; i<n; i++)
        {   UInt08 xi = i;
            q += (xi-avX) * (xi-avX);
        }
    }
    // Steigung auf <period> ausweiten
    w *= n;
    q *= period;
    // Steigung (in Prozent) b = w÷q · 100
    SInt16 b;
    if      (w >  q*10) return -1000; // < -1000 [%]
    else if (w < -q*10) return  1000; // >  1000 [%]
    b = (SInt16)(w*100/q);
    SInt32 remainder = (w*100)%q;
    if (remainder<0) remainder = -remainder;
    if ((q&1) == 0) remainder++;
    if (remainder > q/2) (w<0) ? b-- : b++;
    return -b;  // Negiert, weil in <buffer> die Werte rückwärts eingetragen sind
}

//////////////////////////////////////////////////////////////////////////////////////////



